@@ -5,7 +5,7 @@
  * Licensed under the MIT license. See LICENSE file in the project root for details.
  *
  * Description:
- * Base (generic) implementation of arithmetic on infinite polynomials over GF(2)
+ * Implementation of arithmetic on unlimited polynomials over GF(2)
  *
  *******************************************************************************/
 
@@ -18,16 +18,20 @@
 #if defined(__GNUC__) && defined(__PCLMUL__)
 #include <immintrin.h>
 #define PYGF2X_USE_SSE_CLMUL
+#define KARATSUBA_LIMIT 16
 #elif defined(__GNUC__) && defined(__ARM_NEON__)
 #include <arm_neon.h>
 #define PYGF2X_USE_ARMV7_NEON
+#define KARATSUBA_LIMIT 8
+#else
+#define KARATSUBA_LIMIT 4
 #endif
 
-// Max input value size
+// Max value size
 #define PYGF2X_MAX_DIGITS (9000000/PyLong_SHIFT)
 
 // Define to enable DBG_ASSERT checks
-#define DEBUG_PYGF2X
+//#define DEBUG_PYGF2X
 
 // Define to enable DBG_PRINTF and DBG_PRINTF_DIGITS printouts
 //#define DEBUG_PYGF2X_VERBOSE
@@ -42,9 +46,13 @@
 #endif
 
 #ifdef DEBUG_PYGF2X
-#include <execinfo.h>
 #define DBG_ASSERT(x) { if (! (x) ) { fprintf(stderr,"Assertion failed on line %d: %s\n",__LINE__,#x); my_abort(); } }
+#ifdef __GNUC__
+#include <execinfo.h>
+#endif
+#include <stdlib.h>
 static void my_abort() {
+#ifdef __GNUC__
   void *array[3];
   size_t size;
 
@@ -53,6 +61,7 @@ static void my_abort() {
 
   // print out all the frames to stderr
   backtrace_symbols_fd(array, size, STDERR_FILENO);
+#endif
   exit(1);
 }
 #else
@@ -308,25 +317,29 @@ static void mul_5_nr(digit * restrict const p,
 //
 {
     DBG_ASSERT(l<(1<<5));
-    for(int id_r=0; id_r<nr; id_r++) {
 #if (PyLong_SHIFT == 15)
 #if defined(PYGF2X_USE_SSE_CLMUL)
-	__m128i li = {l,0};
+    __m128i li = {l,0};
+    for(int id_r=0; id_r<nr; id_r++) {
 	__m128i ri = {r0[id_r],0};
 	__m128i pi128 = _mm_clmulepi64_si128(li,ri,0);
 	twodigits pi = pi128[0];
 	p[id_r] ^= pi&PyLong_MASK;
 	p[id_r+1] ^= (pi>>PyLong_SHIFT);
+    }
 #elif defined(PYGF2X_USE_ARMV7_NEON)
+    poly8x8_t li  = {l, l};
+    for(int id_r=0; id_r<nr; id_r++) {
 	poly8x8_t ri  = {r0[id_r], r0[id_r]>> 8 };
-	poly8x8_t li  = {l, l};
 	poly16x8_t  pi128 = vmull_p8(li,ri);
 	uint32_t  pi   = pi128[1];
 	pi <<= 8;
 	pi ^= pi128[0];
 	p[id_r] ^= pi&PyLong_MASK;
 	p[id_r+1] ^= (pi>>PyLong_SHIFT);
+    }
 #else
+    for(int id_r=0; id_r<nr; id_r++) {
 	uint16_t ri = r0[id_r];
 	uint16_t pi_0 = mul_5_5[l][ri&0x1f];
 	ri >>= 5;
@@ -335,18 +348,22 @@ static void mul_5_nr(digit * restrict const p,
 	uint16_t pi_2 = mul_5_5[l][ri&0x1f];
 	p[id_r] ^= ((((pi_2 << 5) ^ pi_1) << 5) ^ pi_0) & PyLong_MASK;
 	p[id_r+1] ^= pi_2 >> 5;
+    }
 #endif
 #elif (PyLong_SHIFT == 30)
 #if defined(PYGF2X_USE_SSE_CLMUL)
-	__m128i li = {l,0};
+    __m128i li = {l,0};
+    for(int id_r=0; id_r<nr; id_r++) {
 	__m128i ri = {r0[id_r],0};
 	__m128i pi128 = _mm_clmulepi64_si128(li,ri,0);
 	twodigits pi = pi128[0];
 	p[id_r] ^= pi&PyLong_MASK;
 	p[id_r+1] ^= (pi>>PyLong_SHIFT);
+    }
 #elif defined(PYGF2X_USE_ARMV7_NEON)
+    poly8x8_t li  = {l, l, l, l};
+    for(int id_r=0; id_r<nr; id_r++) {
 	poly8x8_t ri  = {r0[id_r], r0[id_r]>> 8, r0[id_r]>> 16, r0[id_r]>> 24 };
-	poly8x8_t li  = {l, l, l, l};
         poly16x8_t  pi128 = vmull_p8(li,ri);
 	uint64_t  pi  = pi128[3];
 	pi <<= 8;
@@ -357,7 +374,9 @@ static void mul_5_nr(digit * restrict const p,
 	pi ^= pi128[0];
 	p[id_r] ^= pi&PyLong_MASK;
 	p[id_r+1] ^= (pi>>PyLong_SHIFT);
+    }
 #else
+    for(int id_r=0; id_r<nr; id_r++) {
 	uint32_t ri = r0[id_r];
 	uint16_t pi_0 = mul_5_5[l][ri&0x1f];
 	ri >>= 5;
@@ -372,11 +391,11 @@ static void mul_5_nr(digit * restrict const p,
 	uint16_t pi_5 = mul_5_5[l][ri&0x1f];
 	p[id_r] ^= ((((((((((pi_5 << 5) ^ pi_4) << 5) ^ pi_3) << 5) ^ pi_2) << 5) ^ pi_1) << 5) ^ pi_0) & PyLong_MASK;
 	p[id_r+1] ^= pi_5 >> 5;
+    }
 #endif
 #else
 #error
 #endif
-    }
 }
 
 static void mul_15_nr(digit * restrict const p,
@@ -387,18 +406,20 @@ static void mul_15_nr(digit * restrict const p,
 //
 {
     DBG_ASSERT(l<(1<<15));
-    for(int id_r=0; id_r<nr; id_r++) {
 #if (PyLong_SHIFT == 15)
 #if defined(PYGF2X_USE_SSE_CLMUL)
-	__m128i li = {l,0};
+    __m128i li = {l,0};
+    for(int id_r=0; id_r<nr; id_r++) {
 	__m128i ri = {r0[id_r],0};
 	__m128i pi128 = _mm_clmulepi64_si128(li,ri,0);
 	twodigits pi = pi128[0];
 	p[id_r] ^= pi&PyLong_MASK;
 	p[id_r+1] ^= (pi>>PyLong_SHIFT);
+    }
 #elif defined(PYGF2X_USE_ARMV7_NEON)
+    poly8x8_t li  = {l, l, l>>8, l>>8};
+    for(int id_r=0; id_r<nr; id_r++) {
 	poly8x8_t ri  = {r0[id_r], r0[id_r]>> 8, r0[id_r], r0[id_r]>> 8 };
-	poly8x8_t li  = {l, l, l>>8, l>>8};
 	poly16x8_t  pi128 = vmull_p8(li,ri);
 	uint32_t  pi   = pi128[3];
 	pi <<= 8;
@@ -407,23 +428,29 @@ static void mul_15_nr(digit * restrict const p,
 	pi ^= pi128[0];
 	p[id_r] ^= pi&PyLong_MASK;
 	p[id_r+1] ^= (pi>>PyLong_SHIFT);
+    }
 #else
+    for(int id_r=0; id_r<nr; id_r++) {
 	uint32_t pi = mul_15_15(l, r0[id_r]);
 	p[id_r] ^= (pi & PyLong_MASK);
 	p[id_r+1] ^= (pi >> PyLong_SHIFT);
+    }
 #endif
 #elif PyLong_SHIFT == 30
 #if defined(PYGF2X_USE_SSE_CLMUL)
-	__m128i li = {l,0};
+    __m128i li = {l,0};
+    for(int id_r=0; id_r<nr; id_r++) {
 	__m128i ri = {r0[id_r],0};
 	__m128i pi128 = _mm_clmulepi64_si128(li,ri,0);
 	twodigits pi = pi128[0];
 	p[id_r] ^= pi&PyLong_MASK;
 	p[id_r+1] ^= (pi>>PyLong_SHIFT);
+    }
 #elif defined(PYGF2X_USE_ARMV7_NEON)
+    poly8x8_t li  = {l, l, l, l, l>>8, l>>8, l>>8, l>>8};
+    for(int id_r=0; id_r<nr; id_r++) {
 	poly8x8_t ri  = {r0[id_r], r0[id_r]>> 8, r0[id_r]>> 16, r0[id_r]>> 24,
 			 r0[id_r], r0[id_r]>> 8, r0[id_r]>> 16, r0[id_r]>> 24 };
-	poly8x8_t li  = {l, l, l, l, l>>8, l>>8, l>>8, l>>8};
         poly16x8_t  pi128 = vmull_p8(li,ri);
 	uint64_t  pi  = pi128[7];
 	pi <<= 8;
@@ -436,18 +463,20 @@ static void mul_15_nr(digit * restrict const p,
 	pi ^= pi128[0];
 	p[id_r] ^= pi&PyLong_MASK;
 	p[id_r+1] ^= (pi>>PyLong_SHIFT);
+    }
 #else
+    for(int id_r=0; id_r<nr; id_r++) {
 	uint16_t ri_0 = r0[id_r]&0x7fff;
 	uint16_t ri_1 = r0[id_r]>>15;
 	uint32_t pi_0 = mul_15_15(l, ri_0);
 	uint32_t pi_1 = mul_15_15(l, ri_1);
 	p[id_r] ^= pi_0 ^ ((pi_1 & ((1<<15)-1))<<15);
 	p[id_r+1] ^= (pi_1 >> 15);
+    }
 #endif
 #else
 #error
 #endif
-    }
 }
 
 #if (PyLong_SHIFT == 30)
@@ -455,16 +484,54 @@ static void mul_30_nr(digit * restrict const p,
 		      uint32_t l,
 		      const digit * restrict const r0, int nr)
 //
-// Multiply a bignum polynomial by a 15-bit polynomial
+// Multiply a bignum polynomial by a 30-bit polynomial
 //
 {
     DBG_ASSERT(l<(1<<30));
+#if defined(PYGF2X_USE_SSE_CLMUL)
+    __m128i li = {l,0};
+    for(int id_r=0; id_r<nr; id_r++) {
+	__m128i ri = {r0[id_r],0};
+	__m128i pi128 = _mm_clmulepi64_si128(li,ri,0);
+	twodigits pi = pi128[0];
+	p[id_r] ^= pi&PyLong_MASK;
+	p[id_r+1] ^= (pi>>PyLong_SHIFT);
+    }
+#elif defined(PYGF2X_USE_ARMV7_NEON)
+    poly8x8_t li1 = {l>>16, l>>16, l>>16, l>>16, l>>24, l>>24, l>>24, l>>24};
+    poly8x8_t li0 = {l    , l    , l    , l    , l>> 8, l>> 8, l>> 8, l>> 8};
+    for(int id_r=0; id_r<nr; id_r++) {
+	poly8x8_t ri  = {r0[id_r], r0[id_r]>> 8, r0[id_r]>> 16, r0[id_r]>> 24,
+			 r0[id_r], r0[id_r]>> 8, r0[id_r]>> 16, r0[id_r]>> 24 };
+	poly16x8_t  pi1 = vmull_p8(li1,ri);
+	poly16x8_t  pi0 = vmull_p8(li0,ri);
+	uint64_t  pi  = pi1[7];
+	pi <<= 8;
+	pi ^= pi1[6] ^ pi1[3];
+	pi <<= 8;
+	pi ^= pi1[5] ^ pi1[2] ^ pi0[7];
+	pi <<= 8;
+	pi ^= pi1[4] ^ pi1[1] ^ pi0[6] ^ pi0[3];
+	pi <<= 8;
+	pi ^= pi1[0] ^ pi0[5] ^ pi0[2];
+	pi <<= 8;
+	pi ^= pi0[4] ^ pi0[1];
+	pi <<= 8;
+	pi ^= pi0[0];
+
+	p[id_r  ] ^= pi&PyLong_MASK;
+	p[id_r+1] ^= (pi>>PyLong_SHIFT);
+    }
+
+    return p;
+#else
     for(int id_r=0; id_r<nr; id_r++) {
 	
 	twodigits pi = mul_30_30(l, r0[id_r]);
 	p[id_r] ^= (pi & PyLong_MASK);
 	p[id_r+1] ^= (pi >> PyLong_SHIFT);
     }
+#endif
 }
 #elif !(PyLong_SHIFT == 15)
 #error
@@ -621,6 +688,17 @@ static void mul_nl_nr(digit * restrict const p,
 #endif
 	else
 	    mul_digit_nr(p, r0[0], l0, nl);
+    } else if(nr < KARATSUBA_LIMIT && nl < KARATSUBA_LIMIT) {
+    // Perform standard multiplication
+	twodigits pi = 0;
+	for(int ip=0; ip<nl+nr-1; ip++) {
+	    for(int il=GF2X_MAX(0, ip-nr+1), ir=ip-il; il<nl && ir>=0; il++, ir--) {
+		pi ^= mul_digit_digit(l0[il], r0[ir]);
+	    }
+	    p[ip] ^= pi & PyLong_MASK;
+	    pi >>= PyLong_SHIFT;
+	}
+	p[nl+nr-1] ^= pi;
     } else if(nl > 2*nr) {
     // Divide l to form more equal sized pieces
 	int nc = nl/nr; // Number of chunks
@@ -1093,7 +1171,7 @@ pygf2x_divmod(PyObject *self, PyObject *args)
     int ndigs_u = (nbits_u + (PyLong_SHIFT-1))/PyLong_SHIFT;
     
     int nbits_q = nbits_u > nbits_d-1 ? nbits_u - (nbits_d-1) : 0;
-    int nbits_r = nbits_u > nbits_d-1 ? nbits_u : nbits_d-1; // The final n.o. remainder bits is <= this value.
+    int nbits_r = nbits_u > nbits_d-1 ? nbits_u : nbits_d-1; // Enough room to store u initially, and r finally
     int ndigs_q = (nbits_q + (PyLong_SHIFT-1))/PyLong_SHIFT;
     int ndigs_r = (nbits_r + (PyLong_SHIFT-1))/PyLong_SHIFT;
 
@@ -1109,7 +1187,7 @@ pygf2x_divmod(PyObject *self, PyObject *args)
     DBG_PRINTF("Numerator bits   = %-4d\n",nbits_u);
     DBG_PRINTF("Denominator bits = %-4d\n",nbits_d);
     DBG_PRINTF("Quotient bits    = %-4d\n",nbits_q);
-    DBG_PRINTF("Remainder bits  <= %-4d\n",nbits_r);
+    DBG_PRINTF("Remainder bits  <= %-4d\n",nbits_d-1);
 
     DBG_PRINTF_DIGITS("Numerator        :",numerator->ob_digit,ndigs_u);
     DBG_PRINTF_DIGITS("Denominator      :",denominator->ob_digit,ndigs_d);
@@ -1163,13 +1241,14 @@ pygf2x_divmod(PyObject *self, PyObject *args)
 	     * each step above.
 	     *
 	     * In the implementation below we use |e| which is an entire Python Digit, to make implementation
-	     * simpler and faster for bit polynomials.
+	     * simpler and faster for large polynomials.
+	     *
 	     */
 	    
 	    // Choose accuracy of inverse to compute:
 	    // If nbits_d >= nbits_q just compute q with one single step in the Euclidean division loop below
 	    // Otherwise take multiple steps, each of size nbits_d or less
-	    int nbits_e = GF2X_MIN(nbits_q +1, nbits_d);
+	    int nbits_e = GF2X_MIN(nbits_q, nbits_d);
 	    // Round up to nearest digit size
 	    int ndigs_e = (nbits_e + (PyLong_SHIFT-1))/PyLong_SHIFT;
 	    nbits_e = PyLong_SHIFT*ndigs_e;
@@ -1196,6 +1275,7 @@ pygf2x_divmod(PyObject *self, PyObject *args)
 
 		// dq = ((r >> (nr-ne)) *e) >> (ne-1)
 		digit dq = mul_digit_digit(ei, ri) >> (nbits_ei -1);
+		DBG_PRINTF("ei=%x, ri=%x, dq=%x\n",ei,ri,dq);
 		DBG_ASSERT(dq >= (1u<<(nbits_ei-1)) && dq<(1u<<nbits_ei));
 		// |dq| = nbits_ei
 		q_digits[ndigs_q-1] = dq;
@@ -1207,7 +1287,7 @@ pygf2x_divmod(PyObject *self, PyObject *args)
 		// dr = (dq*d) << nqi
 		digit dr[ndigs_d+1];
 		memset(dr,0,sizeof(dr));
-		mul_digit_nr(dr, dq, denominator->ob_digit, ndigs_d);
+		mul_nl_nr(dr, &dq, 1, denominator->ob_digit, ndigs_d);
 		DBG_PRINTF_DIGITS("dr  :",dr,ndigs_d+1);
 		DBG_PRINTF("dq=%x\n",dq);
 		DBG_ASSERT(ndigs_r -1 - ndigs_qi < ndigs_d +1);
@@ -1237,7 +1317,7 @@ pygf2x_divmod(PyObject *self, PyObject *args)
 		    for(int i=0; i<ndigs_ei; i++)
 			dr[i] = (r_digits[ndigs_ri - ndigs_ei +i] << (PyLong_SHIFT - nbits_ri) & PyLong_MASK) |
 			    r_digits[ndigs_ri - ndigs_ei +i -1] >> nbits_ri;
-		    DBG_PRINTF_DIGITS("r>>(nr-ne)       :",dr,ndigs_ei);		
+		    DBG_PRINTF_DIGITS("r>>(nr-ne)       :",dr,ndigs_ei);
 		    mul_nl_nr(dq, &e[ndigs_e - ndigs_ei], ndigs_ei, dr, ndigs_ei);
 		}
 		rshift(dq, 2*ndigs_ei, nbits_ei-1);
